import requests
import json
import argparse
import psutil
from datetime import datetime, timedelta


def rpcquery(method, rpchost, rpcuser, rpcpassword, params=False):
    '''
        Wrapper to run DefiChain RPC commands
    '''
    if not params:
        params = []
    else:
        params = [params]
    headers = {'Content-type': 'application/json'}
    data = {
        "jsonrpc": "1.0",
        "id": "curltest",
        "method": method,
        "params": params
    }

    response = requests.post(rpchost, auth=(rpcuser, rpcpassword), headers=headers, data=json.dumps(data), timeout=1000)
    response.raise_for_status()

    data = response.json()

    if 'result' in data:
        return data['result']
    return data


def checkAreNodesMining(max_lastblock_seconds, rpchost, rpcuser, rpcpassword):
    '''
        Returns a list of tuples (node_id, True|False) where the boolean defines if a block has successfully been checked within MAX_LASTBLOCK_SECONDS
    '''
    # Get mininginfo
    mininginfo = rpcquery('getmininginfo', rpchost, rpcuser, rpcpassword)
    retval = []

    for node in mininginfo['masternodes']:
        lastBlockTime = datetime.strptime(node['lastblockcreationattempt'], "%Y-%m-%dT%H:%M:%SZ")
        now = datetime.utcnow()
        timeDiff = now - lastBlockTime

        if timeDiff.total_seconds() > max_lastblock_seconds:
            ret = False
        else:
            ret = True

        retval.append((node['id'], ret))

    return retval


def reportJson(key, endpoint, data):
    headers = {'x-api-key': key}
    r = requests.post(f'https://api.defichain-masternode-health.com/v1/{endpoint}', headers=headers, json=data)
    r.raise_for_status()

    data = r.json()
    if 'result' in data:
        return data['result']

    return data


def main():
    parser = argparse.ArgumentParser(description='DefiChain Masternode Monitor')
    parser.add_argument('--max-block-seconds', help='Alert if node did not try to calculate hash within max-block-seconds (default: 30 seconds)', default=30)
    parser.add_argument('--rpcuser', help='RPC username')
    parser.add_argument('--rpcpassword', help='RPC password')
    parser.add_argument('--rpchost', help='RPC host (default: http://localhost:8554)', default='http://localhost:8554')
    parser.add_argument('--verbose', action='store_true', help='Prints stats to stdout')
    parser.add_argument('--defi-path', help='Path to your .defi folder. Example: /home/defi/.defi')
    parser.add_argument('--api-key', help='API Key')

    args = parser.parse_args()

    if args.rpcuser is None or args.rpcpassword is None:
        exit('Please specify rpcuser and rpcpassword argument')

    if args.defi_path is None:
        exit('Please specify defi-path argument')

    if args.api_key is None:
        exit('Please specify an api-key argument')

    try:
        checkNodes = checkAreNodesMining(args.max_block_seconds, args.rpchost, args.rpcuser, args.rpcpassword)
        blockcount = rpcquery('getblockcount', args.rpchost, args.rpcuser, args.rpcpassword)
        bestblockhash = rpcquery('getbestblockhash', args.rpchost, args.rpcuser, args.rpcpassword)
        uptime = rpcquery('uptime', args.rpchost, args.rpcuser, args.rpcpassword)
    except requests.exceptions.HTTPError as err:
        raise SystemExit(err)

    loadavg = psutil.getloadavg()[1]
    vmem = psutil.virtual_memory()
    memUsed = vmem.used / 1024**3
    memTotal = vmem.total / 1024**3

    disk = psutil.disk_usage(args.defi_path)
    diskUsed = disk.used / 1024**3
    diskTotal = disk.total / 1024**3

    if args.verbose:
        print('############ mn server analysis ############')
        print('Load Average: {:.2f}\nMemory Total: {:.0f} GB\nMemory Used: {:.0f} GB\nDisk Total: {:.0f} GB\nDisk Used: {:.0f} GB'.format(loadavg, memTotal, memUsed, diskTotal, diskUsed))
        print('############ mn server analysis ############')
        print('############ mn node info ############')
        print('uptime: {:s}\nLocal block height: {:.0f}\nLocal block hash: {:s}'.format(str(timedelta(seconds=uptime)), blockcount, bestblockhash))
        for nodeId, online in checkNodes:
            print('Operator {:s}: {:s}\n'.format(nodeId, "Online" if online else "Offline!"))
        print('############ mn node info ############')

    data = {
        "load_avg": loadavg,
        "hdd_used": diskUsed,
        "hdd_total": diskTotal,
        "ram_used": memUsed,
        "ram_total": memTotal
    }

    data_node_info = {
        "block_height_local": blockcount,
        "local_hash": bestblockhash,
        "node_uptime": uptime
    }

    try:
        reportJson(args.api_key, 'server-stats', data)
        # reportJson(args.api_key, 'node-info', data_node_info)
    except requests.exceptions.HTTPError as err:
        raise SystemExit(err)



from masternode_health.monitor import rpcquery, checkAreNodesMining, reportJson
from unittest import TestCase, mock
from requests.exceptions import HTTPError
from datetime import datetime


class HealthMonitorTest(TestCase):
    def _mock_response(
            self,
            status=200,
            content="CONTENT",
            json_data=None,
            raise_for_status=None):
        """
        since we typically test a bunch of different
        requests calls for a service, we are going to do
        a lot of mock responses, so its usually a good idea
        to have a helper function that builds these things
        """
        mock_resp = mock.Mock()
        # mock raise_for_status call w/optional error
        if raise_for_status:
            mock_resp.raise_for_status.side_effect = raise_for_status
        # set status code and content
        mock_resp.status_code = status
        mock_resp.content = content
        # add json data if provided
        if json_data:
            mock_resp.json = mock.Mock(
                return_value=json_data
            )
        return mock_resp

    @mock.patch('masternode_health.monitor.requests.post')
    def test_rpcquery_ok(self, mock_post):
        data = {
            'result': {"test": "ok"}
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = rpcquery('getminiginfo', '', '', '')
        self.assertEqual(result['test'], 'ok')

    @mock.patch('masternode_health.monitor.requests.post')
    def test_rpcquery_with_param_ok(self, mock_post):
        data = {
            'result': {"test": "ok"}
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = rpcquery('getminiginfo', '', '', '', params={'id': 1})
        self.assertEqual(result['test'], 'ok')

    @mock.patch('masternode_health.monitor.requests.post')
    def test_rpcquery_failed(self, mock_post):
        mock_resp = self._mock_response(status=500, raise_for_status=HTTPError("rpcerror"))
        mock_post.return_value = mock_resp
        self.assertRaises(HTTPError, rpcquery, 'getmininginfo', '', '', '')

    @mock.patch('masternode_health.monitor.requests.post')
    def test_checkAreNodesMining_single_mn_fails(self, mock_post):
        data = {
            "masternodes": [
                {
                    "id": "8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87",
                    "lastblockcreationattempt": "2021-08-23T19:48:21Z",
                },
            ],
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = checkAreNodesMining(30, '', '', '')

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][0], '8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87')
        self.assertEqual(result[0][1], False)

    @mock.patch('masternode_health.monitor.requests.post')
    def test_checkAreNodesMining_single_mn_ok(self, mock_post):
        datenow = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        data = {
            "masternodes": [
                {
                    "id": "8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87",
                    "lastblockcreationattempt": datenow,
                },
            ],
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = checkAreNodesMining(30, '', '', '')

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][0], '8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87')
        self.assertEqual(result[0][1], True)

    @mock.patch('masternode_health.monitor.requests.post')
    def test_checkAreNodesMining_multiple_mn_ok(self, mock_post):
        datenow = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        data = {
            "masternodes": [
                {
                    "id": "8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87",
                    "lastblockcreationattempt": datenow,
                },
                {
                    "id": "2ceb7c9c3bea0bd0e5e4199eca5d0b797d79a0077a9108951faecf715e1e1a57",
                    "lastblockcreationattempt": datenow,
                }
            ],
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = checkAreNodesMining(30, '', '', '')

        self.assertEqual(len(result), 2)
        self.assertEqual(result[0][0], '8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87')
        self.assertEqual(result[0][1], True)
        self.assertEqual(result[1][0], '2ceb7c9c3bea0bd0e5e4199eca5d0b797d79a0077a9108951faecf715e1e1a57')
        self.assertEqual(result[1][1], True)

    @mock.patch('masternode_health.monitor.requests.post')
    def test_checkAreNodesMining_multiple_mn_one_ok(self, mock_post):
        datenow = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        data = {
            "masternodes": [
                {
                    "id": "8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87",
                    "lastblockcreationattempt": datenow,
                },
                {
                    "id": "2ceb7c9c3bea0bd0e5e4199eca5d0b797d79a0077a9108951faecf715e1e1a57",
                    "lastblockcreationattempt": "2021-08-23T19:48:21Z",
                }
            ],
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = checkAreNodesMining(30, '', '', '')

        self.assertEqual(len(result), 2)
        self.assertEqual(result[0][0], '8cb09568143d7bae6822a7a78f91cb907c23fd12dcf986d4d2c8de89457edf87')
        self.assertEqual(result[0][1], True)
        self.assertEqual(result[1][0], '2ceb7c9c3bea0bd0e5e4199eca5d0b797d79a0077a9108951faecf715e1e1a57')
        self.assertEqual(result[1][1], False)

    @mock.patch('masternode_health.monitor.requests.post')
    def test_checkAreNodesMining_empty_ok(self, mock_post):
        data = {
            "masternodes": [],
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = checkAreNodesMining(30, '', '', '')
        self.assertEqual(len(result), 0)

    @mock.patch('masternode_health.monitor.requests.post')
    def test_reportJson_ok(self, mock_post):
        data = {
            'result': {"message": "ok"}
        }

        mock_resp = self._mock_response(status=200, json_data=data)
        mock_post.return_value = mock_resp

        result = reportJson('key', 'endpoint', {})
        self.assertEqual(result['message'], 'ok')

    @mock.patch('masternode_health.monitor.requests.post')
    def test_reportJson_failed(self, mock_post):
        mock_resp = self._mock_response(status=500, raise_for_status=HTTPError("rpcerror"))
        mock_post.return_value = mock_resp
        self.assertRaises(HTTPError, reportJson, 'key', 'endpoint', {})



from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='masternode_health',
    version='0.1.1',
    author='Christian Sandrini',
    author_email='mail@chrissandrini.ch',
    packages=find_packages(include=['masternode_health', 'masternode_health.*']),
    url='http://pypi.python.org/pypi/masternode_health',
    license='LICENSE.md',
    description='This script is designed for collecting server & DeFiChain node information of your system and send them to the DeFiChain Masternode Health API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['psutil'],
    entry_points={
        'console_scripts': ['masternode-health=masternode_health.monitor:main']
    },
)
